from django.db import migrations, models


def forwards(apps, schema_editor):
    try:
        migrations.AddConstraint(
            model_name='blacklistedtoken',
            constraint=models.UniqueConstraint(fields=('username', 'token'), name='unique_username_token'),
        ),
    except Exception:
        # no constraint on mysql, max key is 3072 bytes which is not enough
        pass


class Migration(migrations.Migration):

    dependencies = [
        ('oauth2_authcodeflow', '0002_auto_20210528_1422'),
    ]

    operations = [
        migrations.AlterField(
            model_name='blacklistedtoken',
            name='token',
            field=models.CharField(editable=False, max_length=15000),
        ),
        migrations.RunPython(forwards),
    ]



from django.urls import path

from .conf import (
    constants,
    settings,
)

urlpatterns = [
    path('authenticate', settings.OIDC_VIEW_AUTHENTICATE.as_view(), name=constants.OIDC_URL_AUTHENTICATION_NAME),
    path('callback', settings.OIDC_VIEW_CALLBACK.as_view(), name=constants.OIDC_URL_CALLBACK_NAME),
    path('logout', settings.OIDC_VIEW_LOGOUT.as_view(), name=constants.OIDC_URL_LOGOUT_NAME),
    path('total_logout', settings.OIDC_VIEW_TOTAL_LOGOUT.as_view(), name=constants.OIDC_URL_TOTAL_LOGOUT_NAME),
    path('logout_by_op', settings.OIDC_VIEW_LOGOUT_BY_OP.as_view(), name=constants.OIDC_URL_LOGOUT_BY_OP_NAME),
]



from logging import warning
from time import time
from typing import (
    Dict,
    Optional,
)

from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import AbstractBaseUser
from django.core.exceptions import SuspiciousOperation
from django.urls import reverse
from jose import jwt
from requests import get as request_get
from requests import post as request_post

from .conf import (
    constants,
    settings,
)
from .models import BlacklistedToken
from .utils import OIDCUrlsMixin


class AuthenticationMixin:
    def validate_and_decode_id_token(self, id_token: str, nonce: Optional[str], jwks: Dict) -> Dict:
        header = jwt.get_unverified_header(id_token)
        if 'alg' not in header:
            raise SuspiciousOperation("No alg value found in header")
        algo = header['alg']
        if algo not in settings.OIDC_RP_SIGN_ALGOS_ALLOWED:
            raise SuspiciousOperation(f"The token algorithm {algo} is not allowed ({', '.join(settings.OIDC_RP_SIGN_ALGOS_ALLOWED)})")
        if algo.startswith('RS'):
            if settings.OIDC_RP_IDP_SIGN_KEY:
                key = settings.OIDC_RP_IDP_SIGN_KEY
            else:
                key = jwks.get(header.get('kid'))
                if not key or key.get('alg', algo) != algo:
                    raise SuspiciousOperation(f"No key found matching key id {header.get('kid')} and algorithm {algo}")
        elif algo.startswith('HS'):
            key = settings.OIDC_RP_CLIENT_SECRET
        else:
            raise NotImplementedError(f"Algo {algo} cannot be handled by this authentication backend")
        if isinstance(key, dict):
            secret = key
        else:  # RSA public key (bytes)
            secret = key
        try:
            claims = jwt.decode(
                id_token,
                secret,
                algorithms=settings.OIDC_RP_SIGN_ALGOS_ALLOWED,
                audience=settings.OIDC_RP_CLIENT_ID,
                options={
                    'verify_aud': True,
                    'verify_iss': False,
                    'verify_sub': False,
                    'verify_iat': False,
                    'verify_at_hash': False,
                },
            )
        except jwt.JWTError as e:
            raise SuspiciousOperation("JWT token verification failed: " + str(e))
        if nonce is not None and claims.get('nonce') != nonce:
            raise SuspiciousOperation("JWT Nonce verification failed")
        return claims

    def validate_claims(self, claims):
        for expected in [settings.OIDC_OP_EXPECTED_EMAIL_CLAIM] + list(settings.OIDC_OP_EXPECTED_CLAIMS):
            if expected not in claims:
                raise SuspiciousOperation(f"'{expected}' claim was expected")

    def get_or_create_user(self, request, id_claims: Dict, access_token: str) -> AbstractBaseUser:
        claims = self.get_full_claims(request, id_claims, access_token)
        username = settings.OIDC_DJANGO_USERNAME_FUNC(claims)
        user, _ = self.UserModel.objects.get_or_create(username=username)
        self.update_user(user, claims)
        user.set_unusable_password()
        user.save()
        return user

    def get_full_claims(self, request, id_claims: Dict, access_token: str) -> Dict:
        """access_token is not used here, id_claims is enough"""
        if settings.OIDC_OP_FETCH_USER_INFO and constants.SESSION_OP_USERINFO_URL in request.session and access_token:
            claims = id_claims.copy()
            claims.update(request_get(
                request.session[constants.SESSION_OP_USERINFO_URL],
                headers={'Authorization': f'{settings.OIDC_AUTHORIZATION_HEADER_PREFIX} {access_token}'},
            ).json())
            return claims
        else:
            return id_claims

    def update_user(self, user: AbstractBaseUser, claims: Dict) -> None:
        """udate the django user with data from the claims"""
        if settings.OIDC_EMAIL_CLAIM is None:
            user.email = settings.OIDC_OP_EXPECTED_EMAIL_CLAIM
        elif callable(settings.OIDC_EMAIL_CLAIM):
            user.email = settings.OIDC_EMAIL_CLAIM(claims)
        else:
            user.email = claims.get(settings.OIDC_EMAIL_CLAIM, '')
        if callable(settings.OIDC_FIRSTNAME_CLAIM):
            user.first_name = settings.OIDC_FIRSTNAME_CLAIM(claims)
        else:
            user.first_name = claims.get(settings.OIDC_FIRSTNAME_CLAIM, '')
        if callable(settings.OIDC_LASTNAME_CLAIM):
            user.last_name = settings.OIDC_LASTNAME_CLAIM(claims)
        else:
            user.last_name = claims.get(settings.OIDC_LASTNAME_CLAIM, '')
        user.is_active = True


class AuthenticationBackend(ModelBackend, AuthenticationMixin):
    def __init__(self, *args, **kwargs):
        self.UserModel = get_user_model()

    def authenticate(self, request,
                     use_pkce: bool,
                     code: str,
                     state: Optional[str], nonce: Optional[str],
                     code_verifier: Optional[str],
                     **kwargs) -> AbstractBaseUser:
        """Authenticates users using OpenID Connect Authorization code flow."""
        if not request:
            return None
        try:
            if use_pkce:
                if any((
                    not code,
                    not code_verifier,
                )):
                    raise SuspiciousOperation('code and code_verifier values are required')
            else:
                if any((
                    not code,
                    not state,
                    not nonce,
                )):
                    raise SuspiciousOperation('code, state and nonce values are required')
            params = {
                'grant_type': 'authorization_code',
                'client_id': settings.OIDC_RP_CLIENT_ID,
                'client_secret': settings.OIDC_RP_CLIENT_SECRET,
                'redirect_uri': request.build_absolute_uri(reverse(constants.OIDC_URL_CALLBACK_NAME)),
                'code': code,
            }
            if use_pkce:
                params['code_verifier'] = code_verifier
            resp = request_post(request.session[constants.SESSION_OP_TOKEN_URL], data=params)
            if resp.status_code != 200:
                raise SuspiciousOperation(f"{resp.status_code} {resp.text}")
            result = resp.json()
            id_token = result['id_token']
            access_token = result['access_token']
            expires_in = result.get('expires_in')  # in secs, could be missing
            refresh_token = result.get('refresh_token') if 'offline_access' in settings.OIDC_RP_SCOPES else None
            id_claims = self.validate_and_decode_id_token(id_token, nonce, request.session[constants.SESSION_OP_JWKS])
            self.validate_claims(id_claims)
            now = time()
            if expires_in:
                expires_at = now + expires_in
            else:
                expires_at = id_claims.get('exp', now + settings.OIDC_MIDDLEWARE_SESSION_TIMEOUT_SECONDS)
            request.session[constants.SESSION_ID_TOKEN] = id_token
            request.session[constants.SESSION_ACCESS_TOKEN] = access_token
            request.session[constants.SESSION_ACCESS_EXPIRES_AT] = expires_at
            request.session[constants.SESSION_EXPIRES_AT] = now + settings.OIDC_MIDDLEWARE_SESSION_TIMEOUT_SECONDS
            if refresh_token:
                request.session[constants.SESSION_REFRESH_TOKEN] = refresh_token
            user = self.get_or_create_user(request, id_claims, access_token)
            return user
        except Exception as e:
            warning(e, str(e))
            return None
        finally:
            # be sure the session is in sync
            request.session.save()


class BearerAuthenticationBackend(ModelBackend, AuthenticationMixin, OIDCUrlsMixin):
    def __init__(self, *args, **kwargs):
        self.UserModel = get_user_model()
        self.authorization_prefix = settings.OIDC_AUTHORIZATION_HEADER_PREFIX
        self.oidc_urls = self.get_oidc_urls({})

    def authenticate(self, request, **kwargs) -> AbstractBaseUser:
        """Authenticates users using the Authorization header and previous OIDC Id Token."""
        try:
            prefix, id_token = request.headers.get('Authorization', ' ').split(' ', 1)
        except ValueError:
            prefix = id_token = ''
        if not prefix or not id_token:
            return None
        try:
            if prefix != self.authorization_prefix:
                raise SuspiciousOperation(f"Authorization should start with a {self.authorization_prefix} prefix")
            if BlacklistedToken.is_blacklisted(id_token):
                raise SuspiciousOperation(f"token {id_token} is blacklisted")
            id_claims = self.validate_and_decode_id_token(id_token, nonce=None, jwks=self.oidc_urls.get(constants.SESSION_OP_JWKS, []))
            user = self.get_or_create_user(request, id_claims, None)
            return user
        except Exception as e:
            warning(e.args[0])
            return None



from typing import Dict

from django.core.exceptions import ImproperlyConfigured
from requests import get as request_get

from .conf import (
    constants,
    settings,
)


class OIDCUrlsMixin:
    def get_oidc_urls(self, session: Dict[str, str]) -> Dict[str, str]:
        session = dict(session.items())  # .copy() is not available on SessionStore
        if settings.OIDC_OP_DISCOVERY_DOCUMENT_URL:
            if any((
                constants.SESSION_OP_AUTHORIZATION_URL not in session,
                constants.SESSION_OP_TOKEN_URL not in session,
                constants.SESSION_OP_USERINFO_URL not in session,
                constants.SESSION_OP_JWKS_URL not in session,
                constants.SESSION_OP_END_SESSION_URL not in session,
            )):
                doc_resp = request_get(settings.OIDC_OP_DISCOVERY_DOCUMENT_URL)
                doc_resp.raise_for_status()
                doc = doc_resp.json()
                session[constants.SESSION_OP_AUTHORIZATION_URL] = doc.get('authorization_endpoint')
                session[constants.SESSION_OP_TOKEN_URL] = doc.get('token_endpoint')
                session[constants.SESSION_OP_USERINFO_URL] = doc.get('userinfo_endpoint')
                session[constants.SESSION_OP_JWKS_URL] = doc.get('jwks_uri')
                session[constants.SESSION_OP_END_SESSION_URL] = doc.get('end_session_endpoint')
        elif any((
            settings.OIDC_OP_AUTHORIZATION_URL,
            settings.OIDC_OP_TOKEN_URL,
            settings.OIDC_OP_USERINFO_URL,
            settings.OIDC_OP_JWKS_URL,
            settings.OIDC_OP_END_SESSION_URL,
        )):
            for conf in ('OP_AUTHORIZATION_URL', 'OP_TOKEN_URL', 'OP_USERINFO_URL', 'OP_JWKS_URL', 'OP_END_SESSION_URL'):
                session_conf = getattr(constants, 'SESSION_' + conf)
                if session_conf not in session:
                    session[session_conf] = getattr(settings, 'OIDC_' + conf)
        confs = ['OP_AUTHORIZATION_URL', 'OP_TOKEN_URL']
        if settings.OIDC_OP_FETCH_USER_INFO:
            confs.append('OP_USERINFO_URL')
        for conf in confs:
            session_conf = getattr(constants, 'SESSION_' + conf)
            if not session.get(session_conf):
                raise ImproperlyConfigured(f'OIDC_{conf} is undefined')
        if not session.get(constants.SESSION_OP_JWKS):
            if session.get(constants.SESSION_OP_JWKS_URL):
                jwks_resp = request_get(session[constants.SESSION_OP_JWKS_URL])
                jwks_resp.raise_for_status()
                jwks = jwks_resp.json()['keys']
            else:
                jwks = []
            session[constants.SESSION_OP_JWKS] = {key['kid']: key for key in jwks if key['kty'] == 'RSA' and key['use'] == 'sig'}
        return session



from django.core.management.base import BaseCommand
from oauth2_authcodeflow.models import BlacklistedToken


class Command(BaseCommand):
    help = "Purge the blacklisted tokens with passed expiry dates"

    def handle(self, *args, **options):
        nb = BlacklistedToken.purge()
        self.stdout.write(f"{nb} blacklisted tokens purged")



from django.core.management.base import (
    BaseCommand,
    CommandError,
)
from django.http import HttpRequest
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from oauth2_authcodeflow.conf import constants


class Command(BaseCommand):
    help = "Show the OIDC url (redirect url, logout url) that should be configured for the application"

    def add_arguments(self, parser):
        parser.add_argument('host', type=str, help="public host name")
        parser.add_argument('--secure', action='store_true', help="specify that HTTPS will be used as public access")

    def handle(self, *args, **options):
        try:
            request = HttpRequest()
            request.META['HTTP_HOST'] = options['host']
            request._get_scheme = lambda: 'https' if options['secure'] else 'http'
            redirect_url = request.build_absolute_uri(reverse(constants.OIDC_URL_CALLBACK_NAME))
            logout_url = request.build_absolute_uri(reverse(constants.OIDC_URL_LOGOUT_BY_OP_NAME))
            self.stdout.write(f"redirect_url: {redirect_url}")
            self.stdout.write(f"logout_url: {logout_url}")
        except NoReverseMatch as e:
            raise CommandError(e)



from base64 import urlsafe_b64encode
from hashlib import sha256
from importlib import import_module
from logging import warning, debug

from django.contrib import auth
from django.core.exceptions import SuspiciousOperation
from django.http import (
    HttpResponse,
    HttpResponseBadRequest,
    HttpResponseRedirect,
)
from django.urls import reverse
from django.utils.crypto import get_random_string
from django.utils.decorators import method_decorator
from django.utils.http import urlencode
from django.views.decorators.cache import never_cache
from django.views.generic import View
from jose import jwt
from urllib.parse import parse_qs

from .conf import (
    constants,
    settings,
)
from .utils import OIDCUrlsMixin
from .models import BlacklistedToken


class CacheBaseView(View, OIDCUrlsMixin):
    @method_decorator(never_cache)
    def dispatch(self, request, *args, **kwargs):
        self._set_cache(request)
        response = super().dispatch(request, *args, **kwargs)
        response['Pragma'] = 'no-cache'  # HTTP 1.0 compat
        response['Expires'] = '0'  # for proxies
        return response

    def _set_cache(self, request):
        for key, value in self.get_oidc_urls(request.session).items():
            request.session[key] = value

    def _clear_cache(self, request):
        for conf in dir(constants):
            if conf.startswith('SESSION') and conf in request.session:
                del request.session[conf]
        request.session.save()


class AuthenticateView(CacheBaseView):
    """
    Ask the OP for a temporary code (auth code flow),
    Using at least the openid scope (OIDC).
    Ends with a redirect.
    """
    http_method_names = ['get']

    def get(self, request):
        from_cli = request.GET.get(constants.OIDC_FROM_CLI_QUERY_STRING)
        if from_cli:
            next_url = '/FROM_CLI_OK'
            failure_url = '/FROM_CLI_FAIL'
            request.session.save()  # ensure request.session.session_key is not None
        else:
            next_url = request.GET.get(settings.OIDC_REDIRECT_OK_FIELD_NAME)
            failure_url = request.GET.get(settings.OIDC_REDIRECT_ERROR_FIELD_NAME)
        if not next_url:
            raise SuspiciousOperation(f"{settings.OIDC_REDIRECT_OK_FIELD_NAME} parameter is required")
        if not failure_url:
            raise SuspiciousOperation(f"{settings.OIDC_REDIRECT_ERROR_FIELD_NAME} parameter is required")
        request.session[constants.SESSION_NEXT_URL] = next_url
        request.session[constants.SESSION_FAIL_URL] = failure_url
        url = request.session[constants.SESSION_OP_AUTHORIZATION_URL]
        exclude_scopes = ['offline_access'] if from_cli else []
        scopes = [scope for scope in settings.OIDC_RP_SCOPES if scope not in exclude_scopes]
        use_pkce = not from_cli and settings.OIDC_RP_USE_PKCE
        auth_params = {
            'response_type': 'code',
            'client_id': settings.OIDC_RP_CLIENT_ID,
            'scope': ' '.join(scopes),
            'redirect_uri': request.build_absolute_uri(reverse(constants.OIDC_URL_CALLBACK_NAME)),
        }
        if 'offline_access' in scopes or settings.OIDC_RP_FORCE_CONSENT_PROMPT:
            auth_params['prompt'] = 'consent'
        if use_pkce:
            code_verifier = get_random_string(length=100, allowed_chars='abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-._~')
            code_challenge = urlsafe_b64encode(sha256(code_verifier.encode('ascii')).digest()).decode('ascii').strip('=')
            auth_params.update({
                'code_challenge_method': 'S256',
                'code_challenge': code_challenge,
            })
            request.session[constants.SESSION_CHALLENGE] = code_verifier
        else:
            if from_cli:
                state = jwt.encode({'session_key': request.session.session_key}, settings.SECRET_KEY, algorithm='HS256')
            else:
                state = get_random_string(settings.OIDC_RANDOM_SIZE)
            nonce = get_random_string(settings.OIDC_RANDOM_SIZE)
            auth_params.update({
                'state': state,
                'nonce': nonce,
            })
            request.session[constants.SESSION_STATE] = state
            request.session[constants.SESSION_NONCE] = nonce
        request.session.save()
        redirect_url = f'{url}?{urlencode(auth_params)}'
        if from_cli:
            # special cases when the request comes from CLI,
            # a redirect will not be usefull, only the redirect location is required.
            return HttpResponse(f"Go to:\n{redirect_url}\n", content_type='text/plain')
        else:
            return HttpResponseRedirect(redirect_url)


class CallbackView(CacheBaseView):
    """
    Callback from the OP.
    Ends with a redirect.
    """
    http_method_names = ['get']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.SessionStore = import_module(settings.SESSION_ENGINE).SessionStore

    def get(self, request):
        debug(f"{request.session.session_key=}, {request.session.keys()=}")
        if 'state' in request.GET:
            state = request.GET['state']
            try:
                jwt.get_unverified_header(state)
                from_cli = True
            except jwt.JWTError:
                from_cli = False
            if from_cli:
                debug("Try to get the session from the state parameter instead of using cookies")
                try:
                    # as called from CLI initialy, the session will not be in the cookie.
                    # hence, load the session from the session_key stored in the jwt state
                    session_key = jwt.decode(state, settings.SECRET_KEY, 'HS256')['session_key']
                    old_session_items = request.session.items()
                    request.session = self.SessionStore(session_key)
                    for key, value in old_session_items:
                        if key not in request.session:
                            request.session[key] = value
                    debug(f"request.session replaced with {request.session.session_key=}")
                except (jwt.JWTError, KeyError):
                    return HttpResponseBadRequest("state appears to be a JWT but the signature failed".encode('utf8'))
        else:
            from_cli = False
        next_url = request.session.get(constants.SESSION_NEXT_URL)
        failure_url = request.session.get(constants.SESSION_FAIL_URL)
        use_pkce = not from_cli and settings.OIDC_RP_USE_PKCE
        if not next_url or not failure_url:
            return HttpResponseBadRequest(f"{constants.SESSION_NEXT_URL} and {constants.SESSION_FAIL_URL} session parameters should be filled".encode('utf8'))
        if request.GET.get('error'):
            warning(request.GET['error'])
            # Make sure the user doesn't get to continue to be logged in in Django
            if request.user.is_authenticated:
                auth.logout(request)
            self._clear_cache(request)
            url = failure_url + '?' + urlencode({'error': request.GET['error']})
        elif all((
            'code' in request.GET,
            settings.OIDC_RP_USE_PKCE or 'state' in request.GET,
        )):
            url = self.auth_callback(request, next_url, failure_url, use_pkce)
        elif all((
            constants.SESSION_LOGOUT_STATE in request.session,
            'state' in request.GET,
        )):
            url = self.logout_callback(request, next_url, failure_url)
        else:
            url = failure_url + '?' + urlencode({'error': 'Unknown OIDC callback'})
        request.session.save()
        if from_cli:
            # special cases when the callback comes from a request initiated from CLI,
            # a redirect will not be usefull, only the header is required or a http 400 with the error message.
            if url.startswith('/FROM_CLI_OK'):
                authorization_prefix = settings.OIDC_AUTHORIZATION_HEADER_PREFIX
                id_token = request.session[constants.SESSION_ID_TOKEN]
                return HttpResponse(f"Header:\n  Authorization: {authorization_prefix} {id_token}\n", content_type='text/plain')
            else:
                error_message = next(iter(parse_qs(url.split('?', 1)[1] if '?' in url else '').get('error', [])), '')
                return HttpResponseBadRequest(f"Error: {error_message}")
        else:
            return HttpResponseRedirect(url)

    def auth_callback(self, request, next_url: str, failure_url: str, use_pkce: bool) -> str:
        url = failure_url
        try:
            code = request.GET['code']
            if use_pkce:
                state = None
                session_nonce = None
                session_challenge = request.session.pop(constants.SESSION_CHALLENGE, None)
                if not session_challenge:
                    raise SuspiciousOperation('OIDC callback: challenge not found in session')
            else:
                session_state = request.session.pop(constants.SESSION_STATE, None)
                if not session_state:
                    raise SuspiciousOperation('OIDC callback: state not found in session')
                state = request.GET['state']
                if state != session_state:
                    raise SuspiciousOperation('OIDC callback: state values do not match')
                session_nonce = request.session.pop(constants.SESSION_NONCE, None)
                if not session_nonce:
                    raise SuspiciousOperation('OIDC callback: nonce not found in session')
                session_challenge = None
            user = auth.authenticate(request, use_pkce=use_pkce, code=code, state=state, nonce=session_nonce, code_verifier=session_challenge)
            if user and user.is_active:
                # keep old session items as auth.login will probably flush the session
                old_session_items = request.session.items()
                auth.login(request, user)
                for key, value in old_session_items:
                    if key not in request.session:
                        request.session[key] = value
                url = next_url
            else:
                url += '?' + urlencode({'error': 'OIDC authent callback, no user error'})
        except Exception as e:
            warning(repr(e))
            url += '?' + urlencode({'error': str(e)})
        return url

    def logout_callback(self, request, next_url: str, failure_url: str) -> str:
        url = failure_url
        try:
            state = request.GET['state']
            session_state = request.session.get(constants.SESSION_LOGOUT_STATE)
            if state == session_state:
                if request.user.is_authenticated:
                    auth.logout(request)
                self._clear_cache(request)
                url = next_url
            else:
                request.session.pop(constants.SESSION_LOGOUT_STATE, None)
                request.session.save()
                url += '?' + urlencode({'error': 'OIDC logout callback, bad state error'})
        except Exception as e:
            msg = e.args[0]
            warning(msg)
            url += '?' + urlencode({'error': str(msg)})
        return url


class LogoutView(CacheBaseView):
    """
    Logout user from the application, called by RP user-agent.
    """
    http_method_names = ['get']

    def get(self, request):
        next_url = request.GET.get(settings.OIDC_REDIRECT_OK_FIELD_NAME)
        failure_url = request.GET.get(settings.OIDC_REDIRECT_ERROR_FIELD_NAME)
        if not next_url:
            return HttpResponseBadRequest(f"{settings.OIDC_REDIRECT_OK_FIELD_NAME} parameter is required".encode('utf8'))
        if not failure_url:
            return HttpResponseBadRequest(f"{settings.OIDC_REDIRECT_ERROR_FIELD_NAME} parameter is required".encode('utf8'))
        try:
            if constants.SESSION_ID_TOKEN not in request.session:
                raise ValueError("id_token is missing from the session, cannot logout")
            id_token = request.session[constants.SESSION_ID_TOKEN]
            return self.logout(request, id_token, next_url, failure_url)
        except Exception as e:
            return HttpResponseRedirect(f"{failure_url}?{urlencode({'error': e.args[0]})}")

    def logout(self, request, id_token, next_url, failure_url):
        BlacklistedToken.blacklist(id_token)
        self._clear_cache(request)
        if request.user.is_authenticated:
            auth.logout(request)
        return HttpResponseRedirect(next_url)


class TotalLogoutView(LogoutView):
    """
    Logout user from the application, the OP and any application connected to the OP , called by RP user-agent.
    """
    def logout(self, request, id_token, next_url, failure_url):
        BlacklistedToken.blacklist(id_token)
        end_session_url = request.session.get(constants.SESSION_OP_END_SESSION_URL)
        if settings.OIDC_OP_TOTAL_LOGOUT and end_session_url:
            state = get_random_string(settings.OIDC_RANDOM_SIZE)
            logout_params = {
                'id_token_hint': id_token,
                'post_logout_redirect_uri': request.build_absolute_uri(reverse(constants.OIDC_URL_CALLBACK_NAME)),
                'state': state,
            }
            self._clear_cache(request)
            request.session[constants.SESSION_NEXT_URL] = next_url
            request.session[constants.SESSION_FAIL_URL] = failure_url
            # id token needs to go back into session because the OP may call the LogoutByOPView on logout
            # and LogoutByOPView needs the id token
            request.session[constants.SESSION_ID_TOKEN] = id_token
            request.session[constants.SESSION_LOGOUT_STATE] = state
            request.session.save()
            redirect_url = f'{end_session_url}?{urlencode(logout_params)}'
            return HttpResponseRedirect(redirect_url)
        else:
            return super().logout(request, id_token, next_url, failure_url)


class LogoutByOPView(CacheBaseView):
    """
    Logout user, called by OP.
    """
    http_method_names = ['get']

    def get(self, request):
        try:
            sid = request.GET.get('sid')
            if not sid:
                raise ValueError("sid parameter is required for logout by OP")
            id_token = request.session.get(constants.SESSION_ID_TOKEN)
            if id_token:
                claims = jwt.get_unverified_claims(id_token)
                if claims.get('sid', sid) != sid:
                    raise ValueError("bad sid parameter")
                BlacklistedToken.blacklist(id_token)
            else:
                raise ValueError("missing id_token session parameter")
            return HttpResponse()
        except Exception as e:
            msg = e.args[0]
            warning(msg)
            return HttpResponseBadRequest(msg.encode('utf8'))



from datetime import (
    datetime,
    timedelta,
)
from logging import warning

from django.contrib.auth import get_user_model
from django.db import models
from django.db.utils import DatabaseError
from django.utils.functional import cached_property
from jose import jwt
from pytz import utc

from .conf import settings


class BlacklistedToken(models.Model):
    username = models.CharField(max_length=255, editable=False, db_index=True)
    # no max length in RFC6749 but:
    # - https://docs.microsoft.com/en-us/linkedin/shared/authentication/programmatic-refresh-tokens
    # - https://stackoverflow.com/questions/24892496/max-size-for-oauth-token
    # postgres, sqlite, mysql >= 5.0.3 or oracle >= 12c required
    token = models.CharField(max_length=15000, editable=False)
    expires_at = models.DateTimeField(db_index=True)
    blacklisted_at = models.DateTimeField(editable=False, db_index=True)

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=['username', 'token'], name='unique_username_token'),
        ]

    @classmethod
    def blacklist(cls, token: str):
        claims = jwt.get_unverified_claims(token)
        username = settings.OIDC_DJANGO_USERNAME_FUNC(claims)
        now = datetime.now(tz=utc)
        expires_at = datetime.fromtimestamp(claims['exp'], tz=utc) if 'exp' in claims else now + timedelta(settings.OIDC_BLACKLIST_TOKEN_TIMEOUT_SECONDS)
        try:
            return cls.objects.create(username=username, token=token, expires_at=expires_at, blacklisted_at=now)
        except DatabaseError as e:
            warning(str(e))
            return None

    @classmethod
    def is_blacklisted(cls, token: str):
        claims = jwt.get_unverified_claims(token)
        username = settings.OIDC_DJANGO_USERNAME_FUNC(claims)
        return cls.objects.filter(username=username, token=token).count() > 0

    @classmethod
    def purge(cls):
        now = datetime.now(tz=utc)
        nb, _ = cls.objects.filter(expires_at__lte=now).delete()
        return nb

    @cached_property
    def user(self):
        User = get_user_model()
        try:
            return User.objects.get(username=self.username)
        except User.DoesNotExist:
            return None

    def __str__(self):
        return f"Blacklisted token for {self.username}, expire at {str(self.expires_at)}"



from logging import debug, error
from re import search
from time import gmtime, strftime, time
from urllib.parse import urlencode

from django.contrib.auth import BACKEND_SESSION_KEY
from django.contrib.sessions.models import Session
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.utils.deprecation import MiddlewareMixin
from django.utils.module_loading import import_string
from requests import post as request_post

from .auth import AuthenticationBackend, BearerAuthenticationBackend
from .conf import constants, settings
from .models import BlacklistedToken


class MiddlewareException(Exception):
    def __str__(self):
        return self.args[0]


class Oauth2MiddlewareMixin(MiddlewareMixin):
    def __init__(self, get_response, token_type, check_function):
        self.token_type = token_type
        self.check_function = check_function
        self.exempt_urls = tuple(
            f'^{reverse(url)}' for url in (
                constants.OIDC_URL_AUTHENTICATION_NAME,
                constants.OIDC_URL_CALLBACK_NAME,
                constants.OIDC_URL_LOGOUT_NAME,
                constants.OIDC_URL_TOTAL_LOGOUT_NAME,
                constants.OIDC_URL_LOGOUT_BY_OP_NAME,
            )
        ) + tuple(settings.OIDC_MIDDLEWARE_NO_AUTH_URL_PATTERNS)
        debug(f"{self.exempt_urls=}")
        super().__init__(get_response)

    def is_oidc_enabled(self, request):
        auth_backend = None
        backend_session = request.session.get(BACKEND_SESSION_KEY)
        if backend_session and request.user and request.user.is_authenticated:
            auth_backend = import_string(backend_session)
        return issubclass(auth_backend, AuthenticationBackend) if auth_backend else False

    def is_refreshable_url(self, request):
        """
        Takes a request and returns whether it triggers a refresh examination
        :arg HttpRequest request:
        :returns: boolean
        """
        # Do not attempt to refresh the session if the OIDC backend is not used
        is_oidc_enabled = self.is_oidc_enabled(request)
        debug(f"{is_oidc_enabled=}, {request.path=}, {self.exempt_urls=}")
        if is_oidc_enabled:
            for url_pattern in self.exempt_urls:
                if search(url_pattern, request.path):
                    return False
            return True
        else:
            return False

    def is_api_request(self, request):
        return any(
            search(url_pattern, request.path)
            for url_pattern in settings.OIDC_MIDDLEWARE_API_URL_PATTERNS
        )

    def process_request(self, request):
        try:
            debug(f"{self=}, {request.session.session_key=}, {request.session.keys()=}")
            if constants.SESSION_ID_TOKEN in request.session:
                id_token = request.session[constants.SESSION_ID_TOKEN]
                if BlacklistedToken.is_blacklisted(id_token):
                    debug(f"token {id_token} is blacklisted")
                    raise MiddlewareException(f"token {id_token} is blacklisted")
            self.check_function(request)
            return
        except MiddlewareException as e:
            if request.method == 'GET':
                next_url = request.GET.get(settings.OIDC_REDIRECT_OK_FIELD_NAME)
                if next_url is None:
                    next_url = request.build_absolute_uri()
                failure_url = request.GET.get(settings.OIDC_REDIRECT_ERROR_FIELD_NAME)
                if failure_url is None:
                    failure_url = request.session.get(constants.SESSION_FAIL_URL, '/')
            else:
                next_url = request.session.get(constants.SESSION_NEXT_URL, '/')
                failure_url = request.session.get(constants.SESSION_FAIL_URL, '/')
            # Destroy session
            Session.objects.get(session_key=request.session.session_key).delete()
            if self.is_api_request(request):
                # Return JSON response
                return JsonResponse({
                    'error': str(e),
                    'token_type': self.token_type,
                }, status=401)
            else:
                # Redirect to authentication page
                return HttpResponseRedirect(reverse(constants.OIDC_URL_AUTHENTICATION_NAME) + '?' + urlencode({
                    settings.OIDC_REDIRECT_OK_FIELD_NAME: next_url,
                    settings.OIDC_REDIRECT_ERROR_FIELD_NAME: failure_url,
                }))


class RefreshAccessTokenMiddleware(Oauth2MiddlewareMixin):
    """
    Refreshes the access token with the OIDC RP after expiry seconds
    For users authenticated with the OIDC RP, verify tokens are still valid and
    if not, force the user to refresh silently.
    """
    def __init__(self, get_response):
        super().__init__(get_response, 'access_token', self.check_access_token)

    def check_access_token(self, request):
        if not self.is_refreshable_url(request):
            debug(f"{request.path} is not refreshable")
            return
        debug(f"{request.path} is refreshable")
        if constants.SESSION_REFRESH_TOKEN not in request.session:
            return
        expiration = request.session[constants.SESSION_ACCESS_EXPIRES_AT]
        now = time()
        if expiration > now:
            # The id_token is still valid, so we don't have to do anything.
            debug('access token is still valid (%s > %s)', strftime('%d/%m/%Y, %H:%M:%S', gmtime(expiration)), strftime('%d/%m/%Y, %H:%M:%S', gmtime(now)))
            return
        debug('access token has expired')
        # The access_token has expired, so we have to refresh silently.
        # Build the parameters.
        params = {
            'grant_type': 'refresh_token',
            'client_id': settings.OIDC_RP_CLIENT_ID,
            'client_secret': settings.OIDC_RP_CLIENT_SECRET,
            'refresh_token': request.session[constants.SESSION_REFRESH_TOKEN],
        }
        resp = request_post(request.session[constants.SESSION_OP_TOKEN_URL], data=params)
        if not resp:
            error(resp.text)
            raise MiddlewareException(resp.text)
        result = resp.json()
        access_token = result['access_token']
        expires_in = result['expires_in']  # in secs
        id_token = result.get('id_token', request.session[constants.SESSION_ID_TOKEN])
        refresh_token = result.get('refresh_token', request.session[constants.SESSION_REFRESH_TOKEN])
        if id_token != request.session[constants.SESSION_ID_TOKEN]:
            # blacklist old token
            BlacklistedToken.blacklist(request.session[constants.SESSION_ID_TOKEN])
        now = time()
        request.session[constants.SESSION_ID_TOKEN] = id_token
        request.session[constants.SESSION_ACCESS_TOKEN] = access_token
        request.session[constants.SESSION_ACCESS_EXPIRES_AT] = now + expires_in
        request.session[constants.SESSION_REFRESH_TOKEN] = refresh_token
        request.session.save()


class RefreshSessionMiddleware(Oauth2MiddlewareMixin):
    """
    Checks if the session expired.
    """
    def __init__(self, get_response):
        if not (10 < settings.OIDC_MIDDLEWARE_SESSION_TIMEOUT_SECONDS < settings.SESSION_COOKIE_AGE):
            raise ImproperlyConfigured("OIDC_MIDDLEWARE_SESSION_TIMEOUT_SECONDS should be less than SESSION_COOKIE_AGE and more than 10 seconds")
        super().__init__(get_response, 'refresh_token', self.check_session)

    def check_session(self, request):
        if not self.is_refreshable_url(request):
            debug(f"{request.path} is not refreshable")
            return
        debug(f"{request.path} is refreshable")
        expiration = request.session.get(constants.SESSION_EXPIRES_AT)
        if not expiration:
            msg = f"No {constants.SESSION_EXPIRES_AT} parameter in the backend session"
            debug(msg)
            raise MiddlewareException(msg)
        now = time()
        if expiration > now:
            # The session is still valid, so we don't have to do anything.
            debug('session is still valid (%s > %s)', strftime('%d/%m/%Y, %H:%M:%S', gmtime(expiration)), strftime('%d/%m/%Y, %H:%M:%S', gmtime(now)))
            return
        # The session has expired, an authentication is now required
        # Blacklist the current id token
        BlacklistedToken.blacklist(request.session[constants.SESSION_ID_TOKEN])
        msg = "Session has expired"
        debug(msg)
        raise MiddlewareException(msg)


class BearerAuthMiddleware(Oauth2MiddlewareMixin):
    """
    Inject User in request if authenticate from header.
    """
    def __init__(self, get_response):
        super().__init__(get_response, None, None)

    def process_request(self, request):
        if 'Authorization' in request.headers:
            user = BearerAuthenticationBackend().authenticate(request)
            if user:
                request.user = user
                if not request.session.session_key:
                    # ensure request.session.session_key exists
                    request.session.save()
        return



from base64 import urlsafe_b64encode
from hashlib import sha1
from typing import (
    Callable,
    Dict,
    Union,
)

from django.conf import settings as dj_settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from . import constants  # noqa F401


def import_string_if_not_func(value: Union[str, Callable], attr: str) -> Callable:
    try:
        return value if callable(value) else import_string(value)
    except ImportError as e:
        raise ImportError(f"Could not import '{value}' for API setting '{attr}'. {e.__class__.__name__}: {e}.")


def get_default_django_username(claims: Dict) -> str:
    """base64 encode of the email hash (sha1)"""
    return urlsafe_b64encode(
        sha1(claims.get('email', '').encode('utf8')).digest()
    ).decode('ascii').rstrip('=')


DEFAULTS = {
    'OIDC_VIEW_AUTHENTICATE': f'{__package__}.views.AuthenticateView',
    'OIDC_VIEW_CALLBACK': f'{__package__}.views.CallbackView',
    'OIDC_VIEW_LOGOUT': f'{__package__}.views.LogoutView',
    'OIDC_VIEW_TOTAL_LOGOUT': f'{__package__}.views.TotalLogoutView',
    'OIDC_VIEW_LOGOUT_BY_OP': f'{__package__}.views.LogoutByOPView',
    # URL of your OpenID connect Provider discovery document url (recommended).
    # If you provide this, the following configs will be ignored:
    # - `OIDC_OP_AUTHORIZATION_URL`
    # - `OIDC_OP_TOKEN_URL`
    # - `OIDC_OP_USERINFO_URL`
    # - `OIDC_OP_JWKS_URL`
    'OIDC_OP_DISCOVERY_DOCUMENT_URL': None,
    # URL of your OpenID connect Provider authorization endpoint (not recommended, `OIDC_OP_DISCOVERY_DOCUMENT_URL` is preferred).
    'OIDC_OP_AUTHORIZATION_URL': None,
    # URL of your OpenID connect Provider token endpoint (not recommended, `OIDC_OP_DISCOVERY_DOCUMENT_URL` is preferred).
    'OIDC_OP_TOKEN_URL': None,
    # URL of your OpenID connect Provider userinfo endpoint (not recommended, `OIDC_OP_DISCOVERY_DOCUMENT_URL` is preferred).
    'OIDC_OP_USERINFO_URL': None,
    # URL of your OpenId connect Provider endpoint to get public signing keys (in `PEM` or `DER` format).
    # This is used to verify the `id_token`.
    # This is not recommended to provide this url here but rather use `OIDC_OP_DISCOVERY_DOCUMENT_URL` config.
    'OIDC_OP_JWKS_URL': None,
    # URL of your OpenID connect Provider end session endpoint (not recommended, `OIDC_OP_DISCOVERY_DOCUMENT_URL` is preferred).
    'OIDC_OP_END_SESSION_URL': None,
    # Fetch user info on login or not.
    'OIDC_OP_FETCH_USER_INFO': True,
    # Do a call to total logout will call the OP for a logout. Default true.
    # Be careful, some OP will not follow the RFC and will not allow the user to NOT logout all connected apps.
    # Azure is such a bad example.
    'OIDC_OP_TOTAL_LOGOUT': True,
    # expected email key.
    'OIDC_OP_EXPECTED_EMAIL_CLAIM': 'email',
    # `OIDC_OP_EXPECTED_EMAIL_CLAIM` value is automatically included in this list.
    'OIDC_OP_EXPECTED_CLAIMS': [],
    # OpenID Connect client ID provided for your Relaying Party/client by your OpenIdConnect Provider
    'OIDC_RP_CLIENT_ID': ImproperlyConfigured,
    # OpenID Connect client secret provided for your Relaying Party/client by your OpenIdConnect Provider
    'OIDC_RP_CLIENT_SECRET': ImproperlyConfigured,
    # `PKCE` improve security, disable it only if your provider cannot handle it.
    'OIDC_RP_USE_PKCE': True,
    # Force to ask for consent on login, even if `offline_access` is not in scopes
    'OIDC_RP_FORCE_CONSENT_PROMPT': False,
    # The OpenID Connect scopes to request during login.
    # The scopes could be usefull later to get access to other ressources.
    # `openid` must be in the list.
    # You can also include the `email` scope to ensure that the email field will be in the claims (recommended).
    # You can also include the `profile` scope to get more (like names, …) info in the `id_token` (recommended).
    # You can also get a `refresh_token` by specifying the `offline_access` scope.
    'OIDC_RP_SCOPES': ['openid', 'email', 'profile', 'offline_access'],
    # Sets the algorithms the IdP may use to sign ID tokens.
    # Typical values ar 'HS256' (no key required) and 'RS256' (public key required)
    # The public keys might be defined in `OIDC_RP_IDP_SIGN_KEY` or deduced using the `OIDC_OP_JWKS_URL` config.
    'OIDC_RP_SIGN_ALGOS_ALLOWED': ['HS256', 'HS384', 'HS512', 'RS256', 'RS384', 'RS512'],
    # Public RSA used to verify signatures. Overrides keys from JWKS endpoint.
    # Should be in PEM or DER format.
    'OIDC_RP_IDP_SIGN_KEY': None,
    # Enables or disables automatic user creation during authentication
    'OIDC_CREATE_USER': True,
    # Sets the length of the random string used in the OAuth2 protocol.
    'OIDC_RANDOM_SIZE': 32,
    # Defines a proxy for all requests to the OpenID Connect provider (fetch JWS, retrieve JWT tokens, Userinfo Endpoint).
    # The default is set to None which means the library will not use a proxy and connect directly.
    # For configuring a proxy check the Python requests documentation: https://requests.readthedocs.io/en/master/user/advanced/#proxies
    'OIDC_PROXY': None,
    # Defines a timeout for all requests to the OpenID Connect provider (fetch JWS, retrieve JWT tokens, Userinfo Endpoint).
    # The default is set to None which means the library will wait indefinitely.
    # The time can be defined as seconds (integer).
    # More information about possible configuration values, see Python requests: https://requests.readthedocs.io/en/master/user/quickstart/#timeouts
    'OIDC_TIMEOUT': None,
    # Sets the GET parameter that is being used to define the redirect URL after succesful authentication
    'OIDC_REDIRECT_OK_FIELD_NAME': 'next',
    # Sets the GET parameter that is being used to define the redirect URL after failed authentication
    'OIDC_REDIRECT_ERROR_FIELD_NAME': 'fail',
    # Function or dotted path to a function that compute the django username based on claims.
    # The username should be unique for this app.
    # The default is to use a base64 encode of the email hash (sha1).
    'OIDC_DJANGO_USERNAME_FUNC': get_default_django_username,
    # Default, by value None, is the value of `OIDC_OP_EXPECTED_EMAIL_CLAIM`
    # You can also provide a lambda that takes all the claims as argument and return an email
    'OIDC_EMAIL_CLAIM': None,
    # You can also provide a lambda that takes all the claims as argument and return a firstname
    'OIDC_FIRSTNAME_CLAIM': 'given_name',
    # You can also provide a lambda that takes all the claims as argument and return a lastname
    'OIDC_LASTNAME_CLAIM': 'family_name',
    'OIDC_BLACKLIST_TOKEN_TIMEOUT_SECONDS': 7 * 86400,  # 7 days
    # Only used when using authorization in header:
    #   Authorization: Bearer id_token
    # This is only possible if oauth2_authcodeflow.auth.BearerAuthenticationBackend has been added to AUTHENTICATION_BACKENDS config list.
    'OIDC_AUTHORIZATION_HEADER_PREFIX': 'Bearer',
    # The RefreshAccessTokenMiddleware and RefreshSessionMiddleware will use this list bypass auth checks.
    # You should include at least any failure/error urls in it.
    'OIDC_MIDDLEWARE_NO_AUTH_URL_PATTERNS': [],
    # The RefreshAccessTokenMiddleware and RefreshSessionMiddleware will use this list to answer JSON response in case of refresh failure.
    # Expected list of regexp URL patterns.
    'OIDC_MIDDLEWARE_API_URL_PATTERNS': ['^/api/'],
    'OIDC_MIDDLEWARE_SESSION_TIMEOUT_SECONDS': 7 * 86400,  # 7 days
}
# settings that may be in dotted string import notation and should be transformed in a lazy way.
IMPORT_STRINGS = [
    'OIDC_VIEW_AUTHENTICATE',
    'OIDC_VIEW_CALLBACK',
    'OIDC_VIEW_LOGOUT',
    'OIDC_VIEW_TOTAL_LOGOUT',
    'OIDC_VIEW_LOGOUT_BY_OP',
    'OIDC_DJANGO_USERNAME_FUNC',
]


class Settings:
    """
    A settings object, that allows settings to be accessed as properties.
    Any setting with dotted string import paths will be resolved to the callable.
    """
    def __init__(self, defaults=None, import_strings=None):
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS

    def __getattr__(self, attr):
        if hasattr(dj_settings, attr):
            val = getattr(dj_settings, attr)
        else:
            if attr not in self.defaults:
                raise ImproperlyConfigured(f"Invalid setting: '{attr}'")
            # Fall back to defaults
            val = self.defaults[attr]
        if val is ImproperlyConfigured:
            raise val(f"Setting '{attr}' not found, it should be defined")
        # Coerce import strings into classes
        if attr in self.import_strings:
            val = import_string_if_not_func(val, attr)
        # Cache the result
        setattr(self, attr, val)
        return val


settings = Settings()



from setuptools import setup, find_packages
from configparser import ConfigParser
from json import loads
# https://github.com/pypa/sampleproject/blob/master/setup.py
# https://packaging.python.org/guides/distributing-packages-using-setuptools


# taken from https://github.com/gsemet/pipenv-to-requirements/blob/master/pipenv_to_requirements/__init__.py
def pkg_clean_version(pkg_name, pkg_info):
    if pkg_info.startswith('{'):
        pkg_info = loads(pkg_info)
    else:
        return pkg_name if pkg_info.strip() == "*" else f"{pkg_name}{pkg_info}"
    if not pkg_info:
        return pkg_name
    version = pkg_info.get("version", "").strip()
    editable = pkg_info.get("editable", False)
    markers = pkg_info["markers"].strip() if pkg_info.get("markers") else ""
    extras = pkg_info.get("extras", [])
    subdir = pkg_info.get("subdirectory", [])
    git = pkg_info.get("git", "").strip()
    path = pkg_info.get("path", "").strip()
    ref = pkg_info.get("ref", "").strip()
    rstr = ""
    if not editable:
        rstr += pkg_name
    if extras:
        rstr += "[{}]".format(', '.join([s.strip() for s in extras]))
    if not editable:
        if version and version != "*":
            rstr += version.strip()
    elif git:
        ref = "@" + ref if ref else ref
        rstr = "-e git+" + git + ref + "#egg=" + pkg_name
        if subdir:
            rstr += '&subdirectory=' + subdir
    else:
        rstr = "-e " + path
    if markers:
        rstr += " ; " + markers
    return rstr


def parse_pip_file(pipfile):
    config = ConfigParser()
    config.read(pipfile)
    return [pkg_clean_version(name.strip('"'), info.strip('"')) for name, info in config._sections.get('packages', {}).items()]


requirements = parse_pip_file('Pipfile')


setup(
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=requirements,
)



# Generated by Django 3.1.1 on 2020-09-29 15:34

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='BlacklistedToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('username', models.CharField(db_index=True, editable=False, max_length=255)),
                ('token', models.TextField(editable=False)),
                ('expires_at', models.DateTimeField(db_index=True)),
                ('blacklisted_at', models.DateTimeField(db_index=True, editable=False)),
            ],
        ),
        # will only be applied if possible
        # migrations.AddConstraint(
        #     model_name='blacklistedtoken',
        #     constraint=models.UniqueConstraint(fields=('username', 'token'), name='unique_username_token'),
        # ),
    ]



for key in [
    'oidc_op_authorization_url',
    'oidc_op_token_url',
    'oidc_op_userinfo_url',
    'oidc_op_jwks_url',
    'oidc_op_jwks',
    'oidc_op_end_session_url',
    'oidc_next_url',
    'oidc_fail_url',
    'oidc_state',
    'oidc_nonce',
    'oidc_challenge',
    'oidc_id_token',
    'oidc_access_token',
    'oidc_access_expires_at',
    'oidc_refresh_token',
    'oidc_expires_at',
    'oidc_logout_state',
]:
    locals()[('session' + key[len('oidc'):]).upper()] = key
OIDC_URL_AUTHENTICATION_NAME = 'oidc_authentication'
OIDC_URL_CALLBACK_NAME = 'oidc_callback'
OIDC_URL_LOGOUT_NAME = 'oidc_logout'
OIDC_URL_TOTAL_LOGOUT_NAME = 'oidc_total_logout'
OIDC_URL_LOGOUT_BY_OP_NAME = 'oidc_logout_by_op'
OIDC_FROM_CLI_QUERY_STRING = 'from_cli'



from django.db import migrations, models


def forwards(apps, schema_editor):
    try:
        migrations.RemoveConstraint(
            model_name='blacklistedtoken',
            name='unique_username_token',
        )
    except Exception:
        # no removal if it’s not there
        pass


class Migration(migrations.Migration):
    dependencies = [
        ('oauth2_authcodeflow', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(forwards),
        migrations.AlterField(
            model_name='blacklistedtoken',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]
